import pygame
from OpenGL.GL import *
from PIL import Image
from pubsub import pub
import uuid

from .style import Style


class Event:
    def __init__(self, value_dict):
        for k, v in value_dict.items():
            setattr(self, k, v)


class Widget:
    DEFAULT_STYLE = Style(color=(1, 1, 1, 1))

    def __init__(self, x=0, y=0, w=0, h=0, style=None, texture=0, *args, **kwargs):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.z = 0
        self.hovered = False
        self.parent = None
        self.elements = []
        if not style:
            style = self.DEFAULT_STYLE
        self.style = style
        self.clicked = None
        self.fade_timer = 0
        self.animation_ratio = 0
        self._color_start = self.style.default_color
        self._color_end = self.style.default_color
        self.texture = texture
        self.cleared = False
        self.direct_rendering = True
        self.dirty = 1
        self.uid = str(uuid.uuid4())
        self._disabled = False
        self.hidden = False

    def disable(self):
        if not self.disabled:
            self.check_enabled()
        prev_disabled = self._disabled
        self.hovered = False
        self.clicked = None
        self._disabled = True
        if not prev_disabled:
            self.reset()

    def check_enabled(self):
        still_disabled = self.disabled
        if still_disabled:
            return
        for element in self.elements:
            element.check_enabled()
        self.reset()

    def enable(self):
        self._disabled = False
        for element in self.elements:
            element.check_enabled()

    def find_element(self, uid):
        if self.uid == uid:
            return self
        for element in self.elements:
            searched = element.find_element(uid)
            if searched:
                return searched
        return None

    def overlaps(self, x, y, w, h):
        s_fbo, fbo_w, fbo_h, s_x, s_y = self.get_draw_parent_fbo()
        rect_1 = pygame.rect.Rect(s_x, s_y, self.w, self.h)
        rect_2 = pygame.rect.Rect(x, y, w, h)
        return rect_1.colliderect(rect_2)

    def clear(self):
        pass

    def set_redraw(self):
        self.cleared = True
        self.dirty = True
        if self.draw_parent:
            self.draw_parent.clear()
            self.draw_parent.set_redraw()

    @property
    def disabled(self):
        if self._disabled:
            return True
        if self.parent:
            return self.parent.disabled
        return False

    @property
    def draw_parent(self):
        if self.parent and self.parent.direct_rendering:
            return self.parent.draw_parent
        return self.parent

    def __repr__(self):
        pres = f"""{self.__class__.__name__}{(self.x, self.y)}"""
        parent = self.parent
        while parent:
            pres = '\t' + pres
            parent = parent.parent
        return pres

    def get_draw_parent_fbo(self):
        if self.parent and self.parent.direct_rendering:
            fbo, w, h, x_parent, y_parent = self.parent.get_draw_parent_fbo()
            return fbo, w, h, x_parent + self.x, y_parent + self.y
        if self.parent:
            return self.parent.fbo, self.parent.total_w, self.parent.total_h, self.x, self.y
        return 0, self.w, self.h, self.x, self.y

    def update(self, frame_time):
        if self.animation_ratio:
            self.animation_ratio = max(0, self.animation_ratio - frame_time / (1000 * self.fade_timer))
            self.clear()
            self.set_redraw()
        for element in self.elements:
            element.update(frame_time)
        if self.dirty and self.draw_parent:
            self.draw_parent.clear()
            self.draw_parent.set_redraw()

    def get_color(self):
        if self.animation_ratio:
            return tuple(self._color_start[i] * self.animation_ratio + \
                         self._color_end[i] * (1 - self.animation_ratio) for i in range(4))
        if self.disabled and self.style.disabled_color:
            return self.style.disabled_color
        if self.clicked and self.style.click_color:
            return self.style.click_color
        if self.hovered and self.style.hover_color:
            return self.style.hover_color
        return self.style.default_color

    def to_element_x(self, x):
        return x - self.x

    def to_element_y(self, y):
        return y - self.y

    def hover_pred(self, x, y):
        return self.x < x < self.x + self.w and self.y < y < self.y + self.h

    def check_mouse(self, x, y):
        if self.disabled:
            return
        for element in self.elements:
            element_x, element_y = self.to_element_x(x), self.to_element_y(y)
            element.check_mouse(element_x, element_y)
        if self.hover_pred(x, y):
            if not self.hovered:
                self.mouse_enter()
        elif self.hovered:
            self.mouse_leave()

    def bind(self, parent):
        self.parent = parent
        if not self.parent.parent:
            self.dirty = 2

    def unbind(self):
        self.parent = None

    def mouse_enter(self):
        redraw = not self.hovered and not self.clicked and self.style.hover_color
        color_start = self.get_color()
        self.hovered = True
        if redraw:
            self._color_start = color_start
            self.fade_timer = self.style.fade_in_time
            self.animation_ratio = 1.0 if self.fade_timer else 0
            self._color_end = self.style.hover_color
            self.dirty = 1
            self.clear()
            self.set_redraw()

    def mouse_leave(self):
        redraw = not self.clicked and self.style.hover_color
        color_start = self.get_color()
        self.hovered = False
        if redraw:
            self._color_start = color_start
            self.fade_timer = self.style.fade_out_time
            self.animation_ratio = 1.0 if self.fade_timer else 0
            self._color_end = self.style.default_color
            self.dirty = 1
            self.clear()
            self.set_redraw()

    def mouse_wheel(self, relative_y):
        pass

    def mouse_down(self, x, y, button):
        if self.disabled:
            return
        if not self.hovered:
            return
        for element in self.elements:
            element.mouse_down(self.to_element_x(x), self.to_element_y(y), button)
        if button != 1:
            return
        if not self.clicked:
            event = Event({'x': x, 'y': y, 'button': button})
            pub.sendMessage(f'{self.uid}.click', event=event)
        redraw = not self.clicked and self.style.click_color
        color_start = self.get_color()
        self.clicked = button
        if redraw:
            self._color_start = color_start
            self.fade_timer = self.style.fade_in_time
            self.animation_ratio = 1.0 if self.fade_timer else 0
            self._color_end = self.style.click_color
            self.dirty = 1
            self.clear()
            self.set_redraw()

    def mouse_up(self, x, y):
        if self.disabled:
            return
        for element in self.elements:
            element.mouse_up(self.to_element_x(x), self.to_element_y(y))
        redraw = self.clicked and self.style.click_color
        color_start = self.get_color()
        if self.clicked and self.hovered:
            event = Event({'x': x, 'y': y, 'button': self.clicked})
            pub.sendMessage(f'{self.uid}.confirm-click', event=event)
        self.clicked = None
        if redraw:
            self._color_start = color_start
            self._color_end = self.style.hover_color if self.hovered else self.style.default_color
            self.fade_timer = self.style.fade_out_time
            self.animation_ratio = 1.0 if self.fade_timer else 0
            self.dirty = 1
            self.clear()
            self.set_redraw()

    def key_down(self, keycode, key_char):
        for element in self.elements:
            element.key_down(keycode, key_char)

    def draw(self, force=False):
        if self.hidden:
            return
        if self.direct_rendering and (self.dirty or force):
            self.parent_draw()
        for element in sorted(self.elements, key=lambda w: w.z):
            element.draw(self.cleared or force and self.direct_rendering)
        if not self.direct_rendering and (self.dirty or force):
            self.parent_draw()
        self.dirty = max(self.dirty - 1, 0)
        self.cleared = False

    def parent_draw(self):
        fbo, w_parent, h_parent, x, y = self.get_draw_parent_fbo()
        if self.texture:
            self.gl_draw_rectangle(self.get_color(), self.texture, fbo, w_parent, h_parent,
                                   off_x=x-self.x, off_y=y-self.y)
        else:
            self.draw_background(fbo, w_parent, h_parent, x, y)

    def draw_background(self, fbo, w_disp, h_disp, x_disp, y_disp):
        if self.style.border_color and self.style.border_line_w:
            self.gl_draw_border(self.style.border_color, 0, fbo, w_disp, h_disp, self.style.border_line_w)
        border_w = self.style.border_line_w
        self.gl_draw_rectangle(self.get_color(), 0, fbo,
                               w_disp, h_disp, border_w+x_disp-self.x,
                               border_w+y_disp-self.y, -2*border_w, -2*border_w)

    def gl_draw_border(self, color, texture, fbo, viewport_w, viewport_h, line_width):
        glBindTexture(GL_TEXTURE_2D, texture)
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        glViewport(0, 0, viewport_w, viewport_h)
        glLineWidth(line_width)
        x, y, w, h = self.x / viewport_w, 1 - (self.y + self.h) / viewport_h, \
                     self.w / viewport_w, self.h / viewport_h

        vert_x1, vert_x2, = x + line_width / viewport_w / 2, x + w - line_width / viewport_w / 2
        v_x1, v_y1, v_x2, v_y2 = -1 + 2 * vert_x1, -1 + 2 * y, -1 + 2 * vert_x2, -1 + (2 * y + 2 * h)
        hori_x, hori_w, = x + line_width / viewport_w, w - line_width / viewport_w * 2
        hori_y1, hori_y2 = y + line_width / viewport_h / 2, y + h - line_width / viewport_h / 2
        h_x1, h_y1, h_x2, h_y2 = -1 + 2 * hori_x, -1 + 2 * hori_y1, -1 + 2 * (hori_x + hori_w), -1 + 2 * hori_y2
        glColor4f(*color)
        glBegin(GL_LINES)
        glVertex2f(v_x1, v_y2)
        glVertex2f(v_x1, v_y1)
        glVertex2f(v_x2, v_y2)
        glVertex2f(v_x2, v_y1)

        glVertex2f(h_x1, h_y1)
        glVertex2f(h_x2, h_y1)
        glVertex2f(h_x1, h_y2)
        glVertex2f(h_x2, h_y2)
        glEnd()

    def gl_draw_rectangle(self, color, texture, fbo, viewport_w, viewport_h, off_x=0, off_y=0, off_w=0, off_h=0,
                          tex_x=0, tex_y=0, tex_w=None, tex_h=None):
        glBindTexture(GL_TEXTURE_2D, texture)
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        glViewport(0, 0, viewport_w, viewport_h)
        x, y, w, h = (self.x + off_x) / viewport_w, 1 - (self.y + self.h + off_y + off_h) / viewport_h, \
                     (self.w + off_w) / viewport_w, (self.h + off_h) / viewport_h
        x1, y1, x2, y2 = -1 + 2 * x, -1 + 2 * y, -1 + 2 * x + 2 * w, -1 + (2 * y + 2 * h)
        tex_x0 = tex_x / (tex_w or self.w)
        tex_y0 = tex_y / (tex_h or self.h)
        tex_x1 = (tex_x + self.w) / (tex_w or self.w)
        tex_y1 = (tex_y + self.h) / (tex_h or self.h)
        glColor4f(*color)
        glBegin(GL_TRIANGLES)
        glTexCoord2f(tex_x0, tex_y0)
        glVertex2f(x1, y1)
        glTexCoord2f(tex_x1, tex_y1)
        glVertex2f(x2, y2)
        glTexCoord2f(tex_x0, tex_y1)

        glVertex2f(x1, y2)
        glTexCoord2f(tex_x0, tex_y0)
        glVertex2f(x1, y1)
        glTexCoord2f(tex_x1, tex_y0)
        glVertex2f(x2, y1)
        glTexCoord2f(tex_x1, tex_y1)
        glVertex2f(x2, y2)
        glEnd()

    def add_element(self, element, position=None):
        if position is not None:
            self.elements.insert(position, element)
        else:
            self.elements.append(element)
        element.bind(self)

    def reset(self):
        self.hovered = False
        self.clicked = None
        self.animation_ratio = 0
        self.clear()
        self.set_redraw()

    def hide(self):
        self.hidden = True
        self.draw_parent.set_redraw()

    def show(self):
        self.hidden = False
        self.draw_parent.set_redraw()

    @classmethod
    def load_image(self, image_path):
        texture = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, texture)
        im = Image.open(image_path)
        im = im.transpose(Image.FLIP_TOP_BOTTOM)
        # Premultiply
        if im.mode == 'RGBA':
            pixels = im.load()
            for px_i in range(im.width):
                for px_j in range(im.height):
                    px_r, px_g, px_b, px_a = pixels[px_i, px_j]
                    pixels[px_i, px_j] = (round(px_r * px_a / 255), round(px_g * px_a / 255),
                                          round(px_b * px_a / 255), px_a)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, im.width, im.height,
                     0, GL_RGBA if im.mode == 'RGBA' else GL_RGB, GL_UNSIGNED_BYTE, im.tobytes())
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        return texture




from OpenGL.GL import *
from .widget import *
from .style import *
from .text import *
from .container import *
from .clickables import *
from .progress_bar import *


def init_gl(width, height):
    # basic opengl configuration
    glViewport(0, 0, width, height)
    glDepthRange(0, 1)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    glClearColor(0.0, 0.0, 0.0, 0.0)
    glClearDepth(1.0)
    glDisable(GL_DEPTH_TEST)
    glDisable(GL_LIGHTING)
    glEnable(GL_BLEND)
    glEnable(GL_TEXTURE_2D)
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)


import pygame
from OpenGL.GL import *
import freetype
import numpy

from .widget import Widget
from .style import Style
from .container import GuiContainer

MAGIC_NUMBER = 32
WRAP_CHAR = 'char'
WRAP_WORDS = 'words'


class RenderFont:
    def __init__(self, font_path, font_size, outline=False, outline_thickness=4):
        self.face = freetype.Face(font_path)
        self.face.set_pixel_sizes(font_size, font_size)
        self.char_to_tex = {}
        self.char_sizes = {}
        self.font_size = font_size
        self.cache_file = font_path[:-4] + f'_{font_size}{"p" if not outline else "o"}.raw'
        self.line_height = 0
        self.min_top = None
        self.outline = outline
        self.outline_thickness = outline_thickness
        self.fill_char_index()
        full_binary = self.get_full_binary()
        self.texture = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, self.texture)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.line_height * MAGIC_NUMBER,
                     len(full_binary) // (MAGIC_NUMBER * self.line_height * 4),
                     0, GL_RGBA, GL_UNSIGNED_BYTE, full_binary)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        self.vbo = Vbo()

    def fill_char_index(self):
        for enum, my_char in enumerate(self.face.get_chars()):
            self.face.load_glyph(my_char[1])
            self.char_to_tex[my_char[1]] = enum
            self.char_sizes[my_char[1]] = (
                self.face.glyph.bitmap_left, self.face.glyph.bitmap_top, self.face.glyph.bitmap.width,
                self.face.glyph.bitmap.rows, self.face.glyph.advance.x >> 6)
            self.line_height = max(self.line_height, self.face.glyph.bitmap.rows)
            self.min_top = min(self.face.glyph.bitmap_top, self.min_top or self.face.glyph.bitmap_top)
        self.line_height -= self.min_top
        self.total_glyph = len(self.char_to_tex) + (-len(self.char_to_tex) % MAGIC_NUMBER)

    def get_full_binary(self):
        try:
            with open(self.cache_file, 'rb') as f:
                return f.read()
        except OSError:
            tex_array = []
            for enum, my_char in enumerate(self.face.get_chars()):
                if self.outline:
                    self.face.load_glyph(my_char[1], freetype.FT_LOAD_FLAGS['FT_LOAD_DEFAULT'] |
                                         freetype.FT_LOAD_FLAGS['FT_LOAD_NO_BITMAP'])
                    glyph = self.face.glyph.get_glyph()
                    stroker = freetype.Stroker()
                    stroker.set(int(self.outline_thickness), freetype.FT_STROKER_LINECAPS['FT_STROKER_LINECAP_ROUND'],
                                freetype.FT_STROKER_LINEJOINS['FT_STROKER_LINEJOIN_ROUND'], 0)
                    glyph.stroke(stroker, True)
                    bitmap = glyph.to_bitmap(freetype.FT_RENDER_MODES['FT_RENDER_MODE_NORMAL'],
                                             freetype.Vector(0, 0), True).bitmap
                else:
                    self.face.load_glyph(my_char[1])
                    bitmap = self.face.glyph.bitmap
                if enum % MAGIC_NUMBER == 0:
                    for i in range(self.line_height):
                        tex_array.append([])
                        for j in range(self.line_height * MAGIC_NUMBER):
                            tex_array[-1].append(bytes([0, 0, 0, 0]))
                start_i = (enum // MAGIC_NUMBER) * self.line_height
                start_j = (enum % MAGIC_NUMBER) * self.line_height
                for i in range(bitmap.width):
                    for j in range(bitmap.rows):
                        tex_array[start_i + j][start_j + i] = bytes(
                            4 * [bitmap.buffer[j * bitmap.width + i]])
            full_binary = b''.join(i for row in tex_array for i in row)
            with open(self.cache_file, 'wb') as f:
                f.write(full_binary)
            return full_binary


class Vbo:
    def __init__(self):
        self.tex_buffer = []
        self.vtx_buffer = []

    def push(self, tex_array, vtx_array):
        self.tex_buffer.extend(tex_array)
        self.vtx_buffer.extend(vtx_array)

    def flush(self, widget):
        if not self.vtx_buffer:
            return
        fbo, w_parent, h_parent, x_disp, y_disp = widget.get_draw_parent_fbo()

        glBindTexture(GL_TEXTURE_2D, widget.texture)
        glBindFramebuffer(GL_FRAMEBUFFER, fbo)
        glViewport(0, 0, w_parent, h_parent)
        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY)
        glColor4f(*widget.get_color())
        glVertexPointer(2, GL_FLOAT, 0, numpy.array(self.vtx_buffer, dtype=numpy.float32).tobytes())
        glTexCoordPointer(2, GL_FLOAT, 0, numpy.array(self.tex_buffer, dtype=numpy.float32).tobytes())
        indices = numpy.array(list(range(len(self.vtx_buffer))), dtype=numpy.uint32)
        glDrawElements(GL_TRIANGLES, len(self.vtx_buffer) // 2, GL_UNSIGNED_INT, indices.tobytes())
        glDisableClientState(GL_VERTEX_ARRAY)
        glDisableClientState(GL_TEXTURE_COORD_ARRAY)
        glFlush()
        self.vtx_buffer = []
        self.tex_buffer = []


class RenderString(Widget):
    DEFAULT_STYLE = Style(color=(1, 1, 1, 1), transparent=True)

    def __init__(self, x, y, string, render_font, wrap=WRAP_WORDS, max_w=None, **kwargs):
        super().__init__(x, y, **kwargs)
        self.max_w = max_w
        self.wrap = wrap
        self.render_font = render_font
        self._string = None
        self._render = []
        self.string = string
        self.texture = render_font.texture
        self.direct_rendering = True
        self.dirty = 1

    @property
    def string(self):
        return self._string

    @string.setter
    def string(self, value):
        if value == self._string:
            return
        self._string = value
        max_x = self.x
        max_y = self.y
        for char, cur_x, cur_y, advance in self.iter_chars():
            max_x = max(max_x, cur_x + advance)
            max_y = cur_y
        self.w, self.h = max_x - self.x, max_y - self.y
        self.dirty = 1
        if self.draw_parent:
            self.draw_parent.clear()
            self.draw_parent.set_redraw()

    def iter_chars(self):
        cur_x, cur_y = self.x, self.y
        chunks = list(self._string) if self.wrap == WRAP_CHAR else self._string.split(' ')
        height = self.render_font.line_height
        for i, chunk in enumerate(chunks):
            if i > 0 and self.wrap == WRAP_WORDS:
                chunk = ' ' + chunk
            size_chunk = sum(
                self.render_font.char_sizes[self.render_font.face.get_char_index(char)][4] for char in chunk)
            if self.max_w is not None and cur_x != self.x and cur_x + size_chunk > self.max_w:
                cur_x = self.x
                cur_y += height
                chunk = chunk.strip()
            if cur_y == self.y:
                cur_y += height + self.render_font.min_top
            for char in chunk:
                if char == "\n":
                    cur_x = self.x
                    cur_y += height
                if char != ' ' and not char.strip():
                    continue
                char = self.render_font.face.get_char_index(char)
                advance = self.render_font.char_sizes[char][4]
                yield char, cur_x, cur_y, advance
                cur_x += advance

    def parent_draw(self):
        fbo, disp_w, disp_h, x_disp, y_disp = self.get_draw_parent_fbo()
        for char, cur_x, cur_y, advance in self.iter_chars():
            rect = pygame.rect.Rect(cur_x - (self.x - x_disp) + self.render_font.char_sizes[char][0],
                                    cur_y - (self.y - y_disp) - self.render_font.char_sizes[char][1],
                                    self.render_font.char_sizes[char][2], self.render_font.char_sizes[char][3])
            x, y = rect.x, rect.y
            x /= disp_w
            y /= disp_h
            w, h = rect.w, rect.h
            tex_i = self.render_font.char_to_tex[char]
            tex_x, tex_y = (tex_i % MAGIC_NUMBER) * self.render_font.line_height, \
                           (tex_i // MAGIC_NUMBER) * self.render_font.line_height
            w, h = self.render_font.line_height / disp_w, self.render_font.line_height / disp_h
            tex_x2, tex_y2 = tex_x + self.render_font.line_height, tex_y + self.render_font.line_height

            mul_x = 1.0 / (MAGIC_NUMBER * self.render_font.line_height)

            mul_y = MAGIC_NUMBER / (self.render_font.total_glyph * self.render_font.line_height)

            tex_x = tex_x * mul_x
            tex_y = tex_y * mul_y
            tex_x2 = tex_x2 * mul_x
            tex_y2 = tex_y2 * mul_y
            vtx_x = -1 + 2 * x
            vtx_y = 1 - 2 * y
            vtx_x2 = -1 + 2 * x + 2 * w
            vtx_y2 = 1 - (2 * y + 2 * h)
            tex_pointer = [tex_x, tex_y, tex_x2, tex_y2, tex_x, tex_y2] + [tex_x, tex_y, tex_x2, tex_y, tex_x2,
                                                                           tex_y2]
            vtx_pointer = [vtx_x, vtx_y, vtx_x2, vtx_y2, vtx_x, vtx_y2] + [vtx_x, vtx_y, vtx_x2, vtx_y, vtx_x2,
                                                                           vtx_y2]
            self.render_font.vbo.push(tex_pointer, vtx_pointer)
        self.render_font.vbo.flush(self)


class TextOverlay(GuiContainer):
    DEFAULT_STYLE = Style(color=(0, 0, 0, 0))

    def __init__(self, x, y, text, font, w=0, h=0, text_style=None, style=None, outline_font=None, **kwargs):
        if outline_font and text_style.border_color:
            border_color, border_w = text_style.border_color, text_style.border_line_w
            text_style.border_color = None
            text_style.border_line_w = 0
            outline_style = Style(color=border_color)
            offset = outline_font.outline_thickness / 64
            self.outline_string = RenderString(-offset, -offset + outline_font.min_top, text, outline_font,
                                               style=outline_style)
        else:
            self.outline_string = None
        self.render_string = RenderString(0, font.min_top, text, font, style=text_style, **kwargs)
        super().__init__(x, y, w or self.render_string.w, h or self.render_string.h, style=style)
        if self.outline_string:
            self.add_element(self.outline_string)
        self.add_element(self.render_string)

    def set_text(self, string):
        if self.outline_string:
            self.outline_string.string = string
        self.render_string.string = string


class TextArea(GuiContainer):
    def __init__(self, x, y, w, h, font, placeholder='', **kwargs):
        super().__init__(x, y, w, h, **kwargs)
        self.string = ''
        self.placeholder = placeholder
        self.render_string = RenderString(0, font.min_top, placeholder, font, max_w=self.w)
        self.add_element(self.render_string)
        self.focus = False

    def key_down(self, keycode, key_char):
        if not self.focus:
            return
        if keycode == pygame.K_BACKSPACE:
            self.string = self.string[:-1]
        else:
            self.string = self.string + key_char
        self.render_string.string = self.string

    def mouse_down(self, x, y, button):
        if button != 1:
            return
        if self.hovered:
            self.focus = True
            if not self.string:
                self.render_string.string = self.string
        else:
            self.focus = False
            if not self.string:
                self.render_string.string = self.placeholder



import pygame
from OpenGL.GL import *

from .widget import Widget
from .style import Style


class OverflowWidget(Widget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.overflow_w = kwargs.get('overflow_w', 0)
        self.overflow_h = kwargs.get('overflow_h', 0)
        self.offset_x = 0
        self.offset_y = kwargs.get('overflow_h', 0)
        if self.overflow_h or self.overflow_w:
            self._scrollbar = ScrollBar(window=self)
            self.add_element(self._scrollbar)
        self.direct_rendering = False

    @property
    def total_w(self):
        return self.w + self.overflow_w

    @property
    def total_h(self):
        return self.h + self.overflow_h

    def mouse_wheel(self, relative_y):
        if self.hovered:
            for element in self.elements:
                element.mouse_wheel(relative_y)
            if self.overflow_h:
                self._scrollbar.y -= 30 * relative_y
                self._scrollbar.scroll()

    def to_element_x(self, x):
        return x - self.x - self.offset_x

    def to_element_y(self, y):
        return y - self.offset_y + self.overflow_h - self.y

    def update(self, frame_time):
        super().update(frame_time)
        if self.dirty and (self.overflow_h or self.overflow_w):
            self._scrollbar.dirty = 1
            self.clear()
            self.set_redraw()

    def reset(self):
        super().reset()
        self.offset_x = 0
        self.offset_y = self.overflow_h
        if self.overflow_h or self.overflow_w:
            self._scrollbar.y = 0


class GuiContainer(OverflowWidget):
    DEFAULT_STYLE = Style(color=(0, 0, 0, 1))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fbo, self.texture = self.create_fbo(self.total_w, self.total_h)
        self.clear()
        self.set_redraw()

    def create_fbo(self, w, h):
        texID = glGenTextures(1)
        if not self.style.transparent:
            glBindTexture(GL_TEXTURE_2D_MULTISAMPLE, texID)
            glTexImage2DMultisample(GL_TEXTURE_2D_MULTISAMPLE, 4, self.get_mode(), w, h, 0)
        else:
            glBindTexture(GL_TEXTURE_2D, texID)
            glTexImage2D(GL_TEXTURE_2D, 0, self.get_mode(), w, h, 0, self.get_mode(), GL_UNSIGNED_BYTE, b'\x22' * 4 * w * h)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        fb_id = glGenFramebuffers(1)
        glBindFramebuffer(GL_FRAMEBUFFER, fb_id)
        glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, texID, 0)
        glDrawBuffers(1, GL_COLOR_ATTACHMENT0)
        return fb_id, texID

    def get_mode(self):
        return GL_RGBA

    def clear(self, cascade=True):
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        glClearColor(*self.get_color())
        glClear(GL_COLOR_BUFFER_BIT)
        if self.style.background:
            x, y, w, h = self.x, self.y, self.w, self.h
            self.x, self.y, self.w, self.h = 0, self.overflow_h, self.total_w, self.total_h
            self.draw_background(self.fbo, self.total_w, self.total_h, 0, 0)
            self.x, self.y, self.w, self.h = x, y, w, h
        super().clear()

    def parent_draw(self):
        fbo, w_disp, h_disp, x_disp, y_disp = self.get_draw_parent_fbo()
        if not self.style.transparent:
            glBindFramebuffer(GL_DRAW_FRAMEBUFFER, fbo)
            glBindFramebuffer(GL_READ_FRAMEBUFFER, self.fbo)
            glBlitFramebuffer(self.offset_x, self.offset_y, self.offset_x + self.w, self.offset_y + self.h, x_disp,
                              (h_disp - y_disp - self.h),
                              x_disp + self.w, h_disp - y_disp, GL_COLOR_BUFFER_BIT, GL_NEAREST)
        else:
            self.gl_draw_rectangle((1, 1, 1, 1), self.texture, fbo, w_disp, h_disp,
                                   off_x=x_disp-self.x, off_y=y_disp-self.y,
                                   tex_w=self.total_w, tex_h=self.total_h,
                                   tex_x=self.offset_x, tex_y=self.offset_y)


class MainWindow(GuiContainer):
    DEFAULT_STYLE = Style(color=(0, 0, 0, 1))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.running = True
        self.direct_rendering = False

    def process_events(self):

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.running = False
            if event.type == pygame.KEYDOWN:
                self.key_down(event.key, event.unicode)
            if event.type == pygame.MOUSEBUTTONDOWN:
                self.mouse_down(*pygame.mouse.get_pos(), event.button)
            if event.type == pygame.MOUSEWHEEL:
                self.mouse_wheel(event.y)
            if event.type == pygame.MOUSEBUTTONUP:
                self.mouse_up(*pygame.mouse.get_pos())
            if event.type == pygame.MOUSEMOTION:
                self.check_mouse(*pygame.mouse.get_pos())


class ScrollBar(Widget):
    DEFAULT_STYLE = Style(color=(1, 1, 1, 0.1), hover_color=(1, 1, 1, 0.2), click_color=(1, 1, 1, 1))

    def __init__(self, window):
        super().__init__(window.w - 8, 0, 8, window.h ** 2 // window.total_h)
        self.drag_start = False
        self.window = window
        self.z = 1
        self.hovered = False

    def hover_pred(self, x, y):
        return self.x < x < self.x + self.w and self.window.overflow_h - self.window.offset_y < y < self.y + self.window.h

    def check_mouse(self, x, y):
        super().check_mouse(x, y)
        if self.clicked:
            self.y += (y - self.y - self.h // 2) * self.window.total_h / (self.window.h - self.h)
            self.scroll()

    def scroll(self):
        self.y = min(max(self.y, 0), self.window.total_h - self.h)
        self.window.offset_y = self.window.overflow_h - self.y * (self.window.overflow_h / (self.window.total_h - self.h))
        self.draw_parent.clear()
        self.draw_parent.set_redraw()

    def mouse_down(self, x, y, button):
        super().mouse_down(x, y, button)
        if button != 1:
            return
        if self.hovered:
            if not self.y < y < self.y + self.h:  # Jump click
                self.y += (y - self.y - self.h//2) * self.window.total_h / (self.window.h - self.h)
                self.scroll()

    def mouse_up(self, x, y):
        super().mouse_up(x, y)
        if not self.hovered:
            self.mouse_leave()


from pubsub import pub

from .widget import Event
from .container import GuiContainer
from .style import Style
from .text import TextOverlay


class Button(GuiContainer):
    def __init__(self, x, y, w, h, text, font, outline_font=None, padding_x=16, padding_y=8, text_style=None, **kwargs):
        self.caption = TextOverlay(padding_x, padding_y, text, font, outline_font=outline_font, text_style=text_style)
        super().__init__(x, y, w or self.caption.w + 2 * padding_x, h or self.caption.h + 2 * padding_y, **kwargs)
        self.add_element(self.caption)


class DropDown(GuiContainer):
    DEFAULT_STYLE = Style(color=(0, 0, 0, 0))

    def __init__(self, x, y, w, h, top_text, option_list, font, **kwargs):
        if 'style' not in kwargs:
            kwargs['style'] = self.DEFAULT_STYLE
        self.button = Button(0, 0, w, h, top_text, font, style=kwargs['style'])
        options_h = 0
        self.options = []
        for i, text in enumerate(option_list):
            queue_args = {}
            option = Button(0, 0, w, h, text, font, style=kwargs['style'], **queue_args)
            option.y = options_h
            options_h += option.h
            self.options.append(option)
        total_h = options_h + self.button.h
        max_h = kwargs.get('max_h', total_h)
        overflow_h = total_h - max_h
        self.drop_down = GuiContainer(0, self.button.h, w, max_h - self.button.h,
                                      style=Style(color=(1, 1, 1, 0)), overflow_h=overflow_h)
        for option in self.options:
            self.drop_down.add_element(option)
        super().__init__(x, y, w, max_h, style=Style(color=(1, 1, 1, 0)))
        self.add_element(self.button)
        self.focus = False

    def mouse_down(self, x, y, button):
        super(DropDown, self).mouse_down(x, y, button)
        if button != 1 or self.drop_down.overflow_h and self.drop_down._scrollbar.clicked:
            return
        if self.hovered and not self.focus:
            pub.sendMessage(f'{self.uid}.focus', event=Event({}))
            self.focus = True
            self.add_element(self.drop_down)
            self.clear()
            self.set_redraw()
        elif self.focus:
            for i, element in enumerate(self.options):
                if element.clicked:
                    pub.sendMessage(f'{self.uid}.select', event=Event({'index': i}))
            self.elements.remove(self.drop_down)
            pub.sendMessage(f'{self.uid}.unfocus', event=Event({}))
            self.focus = False
            self.clear()
            self.set_redraw()
            for option in self.options:
                option.reset()
            self.drop_down.reset()
            self.drop_down.unbind()

    def hover_pred(self, x, y):
        if not self.focus:
            return self.button.hover_pred(self.to_element_x(x), self.to_element_y(y))
        return super().hover_pred(x, y)



from .container import GuiContainer
from .widget import Widget
from .style import Style


class ProgressBar(GuiContainer):
    DEFAULT_FILL_STYLE = Style(color=(0, 0.4, 0.1, 1))

    def __init__(self, x, y, w, h, fill_style=None, **kwargs):
        super().__init__(x, y, w, h, **kwargs)
        self._ratio = 0.0
        self.fill_style = fill_style or self.DEFAULT_FILL_STYLE
        self.inner_rect = None

    def set_progress(self, ratio):
        ratio = min(max(ratio, 0), 1.0)
        border_w = self.style.border_line_w
        new_w = round((self.w - 2 * border_w) * ratio)
        if new_w <= 0:
            if self.inner_rect in self.elements:
                self.elements.remove(self.inner_rect)
            return
        if not self.inner_rect and new_w:
            self.inner_rect = Widget(border_w, border_w, self.w - 2 * border_w, self.h - 2 * border_w,
                                     style=self.fill_style)
        if new_w and self.inner_rect not in self.elements:
            self.add_element(self.inner_rect)
        self._ratio = ratio
        old_w = self.inner_rect.w
        self.inner_rect.w = new_w
        if self.inner_rect.w != old_w:
            self.inner_rect.clear()
            self.inner_rect.set_redraw()





class Style:
    def __init__(self,
                 parent_styles=None,
                 color=(0, 0, 0, 0),
                 hover_color=None,
                 click_color=None,
                 disabled_color=None,
                 border_color=None,
                 border_line_w=0,
                 fade_in_time=0.0,
                 fade_out_time=0.0,
                 transparent=None):
        if parent_styles:
            for parent_style in reversed(parent_styles):
                attrs = parent_style.__dict__
                for k, v in attrs.items():
                    setattr(self, k, v)
        self.default_color = self.premultiply(color)
        self.hover_color = self.premultiply(hover_color)
        self.click_color = self.premultiply(click_color)
        self.disabled_color = self.premultiply(disabled_color)
        self.transparent = transparent if transparent is not None else self.default_color[3] < 1.0
        self.fade_in_time = fade_in_time
        self.fade_out_time = fade_out_time
        self.border_color = border_color
        self.border_line_w = border_line_w

    @property
    def background(self):
        return self.hover_color or self.border_color

    def premultiply(self, color):
        if not color:
            return color
        return color[0] * color[3], color[1] * color[3], color[2] * color[3], color[3]

    def __str__(self):
        return f'#{int(255 * self.default_color[0]):02X}{int(255 * self.default_color[1]):02X}' \
               f'{int(255 * self.default_color[2]):02X}{int(255 * self.default_color[3]):02X}'
